import datetime
from pydantic import BaseModel
from mad_prefect.data_assets import ASSET_METADATA_LOCATION
from mad_prefect.filesystems import get_fs


class DataAssetRun(BaseModel):
    id: str | None = None
    runtime: datetime.datetime | None = None
    materialized: datetime.datetime | None = None
    duration_miliseconds: int | None = None

    asset_id: str | None = None
    asset_name: str | None = None
    asset_path: str | None = None
    parameters: str | None = None

    async def persist(self):
        fs = await get_fs()

        await fs.write_data(
            f"{ASSET_METADATA_LOCATION}/asset_name={self.asset_name}/asset_id={self.asset_id}/asset_run_id={self.id}/metadata.json",
            self,
        )



from typing import cast
import duckdb
from mad_prefect.data_assets import ARTIFACT_FILE_TYPES
from mad_prefect.data_assets.options import ReadCSVOptions, ReadJsonOptions
from mad_prefect.duckdb import register_mad_protocol
from mad_prefect.data_assets.data_artifact import DataArtifact


class DataArtifactQuery:

    def __init__(
        self,
        artifacts: list[DataArtifact] | None = None,
        read_json_options: ReadJsonOptions | None = None,
        read_csv_options: ReadCSVOptions | None = None,
    ):
        self.artifacts = artifacts or []
        self.read_json_options = read_json_options or ReadJsonOptions()
        self.read_csv_options = read_csv_options or ReadCSVOptions()

    async def query(self, query_str: str | None = None):
        await register_mad_protocol()

        # Get the globs for any artifacts which exist
        existing_artifacts = [a for a in self.artifacts if await a.exists()]
        globs = [f"mad://{a.path.strip('/')}" for a in existing_artifacts]

        if not globs:
            return

        # Ensure each artifact is of the same filetype
        filetypes = set([a.filetype for a in existing_artifacts])

        if not filetypes or len(filetypes) > 1:
            raise ValueError("Cannot query artifacts of different filetypes")

        # Get the base query
        filetype: ARTIFACT_FILE_TYPES = cast(ARTIFACT_FILE_TYPES, filetypes.pop())

        if filetype == "json":
            artifact_query = self._create_query_json(globs)
        elif filetype == "parquet":
            artifact_query = self._create_query_parquet(globs)
        elif filetype == "csv":
            artifact_query = self._create_query_csv(globs)
        else:
            raise ValueError(f"Unsupported file format {filetype}")

        # Apply any additional query on top
        if query_str:
            return duckdb.query(f"FROM artifact_query {query_str}")

        return artifact_query

    def _create_query_json(self, globs: list[str]):
        # Prepare the globs string
        globs_str = ", ".join(f"'{g}'" for g in globs)
        globs_formatted = f"[{globs_str}]"

        # Build the base options dict without 'columns'
        base_options = self.read_json_options.model_dump(
            exclude={"columns"},
            exclude_none=True,
        )
        options_str = self._format_options_dict(base_options)

        # Build the base query string without 'columns'
        base_query = (
            f"SELECT * FROM read_json({globs_formatted}, {options_str})"
            if options_str
            else f"SELECT * FROM read_json({globs_formatted})"
        )

        # Process columns after building the base query
        if self.read_json_options.columns:
            updated_columns = self._process_columns(
                base_query, self.read_json_options.columns
            )

            # Include 'columns' in options
            options_with_columns = base_options.copy()
            options_with_columns["columns"] = updated_columns
            options_str_with_columns = self._format_options_dict(options_with_columns)

            # Rebuild the query with 'columns'
            final_query = f"SELECT * FROM read_json({globs_formatted}, {options_str_with_columns})"
        else:
            final_query = base_query

        # Execute the query
        artifact_query = duckdb.query(final_query)
        return artifact_query

    def _process_columns(
        self,
        base_query: str,
        columns: dict[str, str],
    ) -> dict[str, str]:
        # Describe the base query to get the schema
        schema_info = duckdb.query(f"DESCRIBE {base_query}").fetchall()
        schema_columns = {row[0]: row[1] for row in schema_info}

        # Update column types based on provided columns
        updated_columns = {}
        for col_name, col_type in schema_columns.items():
            if col_name in columns:
                # Use the provided type
                updated_columns[col_name] = columns[col_name]
            else:
                # Use the existing type from the schema
                updated_columns[col_name] = col_type

        return updated_columns

    def _create_query_parquet(self, globs: list[str]):
        # Prepare the globs string
        globs_str = ", ".join(f"'{g}'" for g in globs)
        globs_formatted = f"[{globs_str}]"

        # Include only relevant options
        options_dict = {"hive_partitioning": True, "union_by_name": True}
        options_str = self._format_options_dict(options_dict)

        # Build the query string
        artifact_base_query = (
            f"SELECT * FROM read_parquet({globs_formatted}, {options_str})"
        )

        # Execute the query
        artifact_query = duckdb.query(artifact_base_query)
        return artifact_query

    def _create_query_csv(self, globs: list[str]):
        # Convert each artifact path to a DuckDB-friendly string
        globs_str = ", ".join(f"'{g}'" for g in globs)
        globs_formatted = f"[{globs_str}]"

        # Build the base options dict without 'columns'
        base_options = self.read_csv_options.model_dump(
            exclude_none=True,
        )

        options_str = self._format_options_dict(base_options)

        # Build the base query string without 'columns'
        base_query = (
            f"SELECT * FROM read_csv({globs_formatted}, {options_str})"
            if options_str
            else f"SELECT * FROM read_csv({globs_formatted})"
        )

        # Execute the query
        artifact_query = duckdb.query(base_query)
        return artifact_query

    def _format_options_dict(self, options_dict: dict) -> str:
        def format_value(key, value):
            if isinstance(value, bool):
                return "TRUE" if value else "FALSE"
            elif isinstance(value, str):
                return f"'{value}'"
            elif isinstance(value, dict):
                return f"{value}"
            else:
                return str(value)

        options_str = ", ".join(
            f"{key} = {format_value(key, value)}" for key, value in options_dict.items()
        )
        return options_str



from pydantic import BaseModel
from typing import Dict, List, Optional, Union


class ReadJsonOptions(BaseModel):
    hive_partitioning: Optional[bool] = True
    union_by_name: Optional[bool] = True
    maximum_object_size: Optional[int] = 33554432
    format: Optional[str] = "auto"

    # By default, always try and parse a nested object as a struct
    field_appearance_threshold: Optional[int] = 0
    map_inference_threshold: Optional[int] = -1
    sample_size: Optional[int] = -1

    columns: Optional[Dict[str, str]] = None
    timestampformat: Optional[str] = None
    auto_detect: Optional[bool] = None
    compression: Optional[str] = None
    convert_strings_to_integers: Optional[bool] = None
    dateformat: Optional[str] = None
    filename: Optional[bool] = None
    ignore_errors: Optional[bool] = None
    maximum_depth: Optional[int] = None
    records: Optional[str] = None


class ReadCSVOptions(BaseModel):
    hive_partitioning: Optional[bool] = True
    union_by_name: Optional[bool] = True
    header: Optional[bool] = True
    auto_detect: Optional[bool] = True

    # Below are all other DuckDB read_csv parameters,
    # defaulting to None so that DuckDB's built-in defaults apply.
    all_varchar: Optional[bool] = None
    allow_quoted_nulls: Optional[bool] = None
    auto_type_candidates: Optional[List[str]] = None
    buffer_size: Optional[int] = None
    columns: Optional[Dict[str, str]] = None
    comment: Optional[str] = None
    compression: Optional[str] = None
    dateformat: Optional[str] = None
    decimal_separator: Optional[str] = None
    delim: Optional[str] = None
    escape: Optional[str] = None
    encoding: Optional[str] = None
    filename: Optional[bool] = None
    force_not_null: Optional[List[str]] = None
    ignore_errors: Optional[bool] = None
    max_line_size: Optional[int] = None
    column_names: Optional[List[str]] = None
    new_line: Optional[str] = None
    normalize_names: Optional[bool] = None
    null_padding: Optional[bool] = None
    nullstr: Optional[Union[str, List[str]]] = None
    null: Optional[Union[str, List[str]]] = None
    parallel: Optional[bool] = None
    quote: Optional[str] = None
    rejects_scan: Optional[str] = None
    rejects_table: Optional[str] = None
    rejects_limit: Optional[int] = None
    rfc_4180: Optional[bool] = None
    sample_size: Optional[int] = None
    sep: Optional[str] = None
    skip: Optional[int] = None
    store_rejects: Optional[bool] = None
    timestampformat: Optional[str] = None
    timestamp_format: Optional[str] = None
    types: Optional[Union[List[str], Dict[str, str]]] = None
    dtypes: Optional[Union[List[str], Dict[str, str]]] = None
    column_types: Optional[Union[List[str], Dict[str, str]]] = None



from inspect import isasyncgen, iscoroutine, isgenerator


async def yield_data_batches(data: object):
    if iscoroutine(data):
        data = await data

    if isgenerator(data) or isasyncgen(data):
        if isasyncgen(data):
            async for d in data:
                yield d
        else:
            for d in data:
                yield d
    else:
        yield data



from datetime import datetime, UTC, timedelta, timezone
import hashlib
import inspect
import json
import re
from typing import Callable, List
import duckdb
from mad_prefect.data_assets import ARTIFACT_FILE_TYPES
from mad_prefect.data_assets import ASSET_METADATA_LOCATION
from mad_prefect.data_assets.data_artifact import DataArtifact
from mad_prefect.data_assets.data_artifact_collector import DataArtifactCollector
from mad_prefect.data_assets.data_artifact_query import DataArtifactQuery
from mad_prefect.data_assets.data_asset_run import DataAssetRun
from mad_prefect.data_assets.options import ReadCSVOptions, ReadJsonOptions
from mad_prefect.duckdb import register_mad_protocol
from mad_prefect.filesystems import get_fs
import os


class DataAsset:
    def __init__(
        self,
        fn: Callable,
        path: str,
        artifacts_dir: str = "",
        name: str | None = None,
        snapshot_artifacts: bool = False,
        artifact_filetype: ARTIFACT_FILE_TYPES = "json",
        read_json_options: ReadJsonOptions | None = None,
        read_csv_options: ReadCSVOptions | None = None,
        cache_expiration: timedelta | None = None,
    ):
        self._fn: Callable = fn
        self._fn_signature: inspect.Signature = inspect.signature(fn)
        self._bound_arguments: inspect.BoundArguments | None = None

        self.name: str = name if name else fn.__name__
        self.path: str = path
        self.artifacts_dir: str = artifacts_dir
        self.snapshot_artifacts: bool = snapshot_artifacts

        self.artifact_filetype: ARTIFACT_FILE_TYPES = artifact_filetype
        self.cache_expiration: timedelta = cache_expiration or timedelta(0)

        self.id = self._generate_asset_guid()

        self.asset_run = DataAssetRun()
        self.asset_run.id = self._generate_asset_iteration_guid()
        self.asset_run.asset_name = self.name
        self.asset_run.asset_path = self.path

        self.read_json_options = read_json_options or ReadJsonOptions()
        self.read_csv_options = read_csv_options or ReadCSVOptions()

        # If the function has no parameters, bind empty arguments immediately
        if not self._fn_signature.parameters:
            self._bind_arguments()

    def with_arguments(self, *args, **kwargs):
        asset = DataAsset(
            self._fn,
            self.path,
            self.artifacts_dir,
            self.name,
            self.snapshot_artifacts,
            self.artifact_filetype,
            self.read_json_options,
            self.read_csv_options,
            self.cache_expiration,
        )

        asset._bind_arguments(*args, **kwargs)
        return asset

    def with_options(
        self,
        path: str | None = None,
        artifacts_dir: str | None = None,
        name: str | None = None,
        snapshot_artifacts: bool | None = None,
        artifact_filetype: ARTIFACT_FILE_TYPES | None = None,
        read_json_options: ReadJsonOptions | None = None,
        read_csv_options: ReadCSVOptions | None = None,
        cache_expiration: timedelta | None = None,
    ):
        # Default to the current asset's options for any None values
        asset = DataAsset(
            self._fn,
            path=path or self.path,
            artifacts_dir=artifacts_dir or self.artifacts_dir,
            name=name or self.name,
            snapshot_artifacts=snapshot_artifacts or self.snapshot_artifacts,
            artifact_filetype=artifact_filetype or self.artifact_filetype,
            read_json_options=read_json_options or self.read_json_options,
            read_csv_options=read_csv_options or self.read_csv_options,
            cache_expiration=cache_expiration or self.cache_expiration,
        )

        # Ensure we're also passing through any bound arguments if we have them
        if self._bound_arguments:
            asset._bind_arguments(
                *self._bound_arguments.args, **self._bound_arguments.kwargs
            )

        return asset

    def _bind_arguments(self, *args, **kwargs):
        self._bound_arguments = self._fn_signature.bind(*args, **kwargs)
        self._bound_arguments.apply_defaults()

        args = self._bound_arguments
        args_dict = dict(args.arguments)

        # Resolve (path, artifacts_dir, name) to insert template values
        resolved_path = self._resolve_attribute(self.path)
        resolved_name = self._resolve_attribute(self.name)
        resolved_artifacts_dir = self._resolve_attribute(self.artifacts_dir) or ""

        if not resolved_path:
            raise ValueError(
                f"Unable to resolve data asset path: {self.path} with {args_dict}."
            )

        if not resolved_name:
            raise ValueError(
                f"Unable to resolve data asset name: {self.name} with {args_dict}."
            )

        self.path = resolved_path
        self.name = self._sanitize_name(resolved_name)
        self.artifacts_dir = resolved_artifacts_dir

        def _handle_unknown_types(data):
            if isinstance(data, DataAsset):
                return {"name": self.name, "fn": self._fn}

        # Recalculate the ids in case the parameters have changed
        self.id = self.asset_run.asset_id = self._generate_asset_guid()
        self.asset_run.id = self._generate_asset_iteration_guid()
        self.asset_run.asset_path = self.path
        self.asset_run.asset_name = self.name
        self.asset_run.parameters = json.dumps(args_dict, default=_handle_unknown_types)

        return self

    async def __call__(self, *args, **kwargs):
        if args or kwargs:
            # If arguments are passed in, create a new instance with bound arguments
            asset_with_arguments = self.with_arguments(*args, **kwargs)
            return await asset_with_arguments()
        else:
            # No arguments passed in
            if not self._bound_arguments:
                # If the function expects no parameters, bind empty arguments
                if not self._fn_signature.parameters:
                    self._bind_arguments()
                else:
                    raise TypeError(
                        f"{self.name}() missing required positional arguments"
                    )

        assert self._bound_arguments

        # Extract and set filetypes for result artifacts
        self.result_artifact_filetypes = self.get_result_artifact_filetypes()

        # There may be multiple result artifacts if following syntax is used "bronze/customers.parquet|csv"
        self.result_artifacts = self._create_result_artifacts()

        # Prevent asset from being rematerialized inside a single session
        if self.asset_run and (materialized := self.asset_run.materialized):
            return self.result_artifacts[0]

        self.asset_run.runtime = datetime.now(UTC)
        self.last_materialized = await self._get_last_materialized()

        # If data has been materialized within cache_expiration period return empty result_artifact
        if await self._cached_result(self.result_artifacts[0], self.asset_run.runtime):
            return self.result_artifacts[0]

        # Regenerate asset_run_id as runtime has now been set.
        self.asset_run.id = self._generate_asset_iteration_guid()

        print(
            f"Running operations for asset_run_id: {self.asset_run.id}, on asset_id: {self.id}, on asset: {self.name}"
        )

        # Write metadata before processing result for troubleshooting purposes
        await self.asset_run.persist()

        # For each fragment in the data batch, we create a new artifact
        base_artifact_path = self._get_artifact_base_path()

        # Clean up the old directory and delete it if we're not snapshotting
        if not self.snapshot_artifacts:
            fs = await get_fs()
            await fs.delete_path(base_artifact_path, recursive=True)

        collector = DataArtifactCollector(
            self._fn(*self._bound_arguments.args, **self._bound_arguments.kwargs),
            base_artifact_path,
            self.artifact_filetype,
            read_json_options=self.read_json_options,
            read_csv_options=self.read_csv_options,
        )

        # Collect the artifacts yielded from the materialization fn
        collector_artifacts = await collector.collect()

        # Query each of the artifacts [filepath1, filepath2, etc] with duckdb
        artifact_query = DataArtifactQuery(
            artifacts=collector_artifacts,
            read_json_options=self.read_json_options,
            read_csv_options=self.read_csv_options,
        )

        # The result is all the artifacts unioned
        result_artifact_data = await artifact_query.query()

        for result_artifact in self.result_artifacts:
            result_artifact.data = result_artifact_data

            # Persist the result artifact to storage, fully materialize it
            await result_artifact.persist()

        # Release reference to data
        result_artifact_data = None

        # Record information about the run
        self.asset_run.materialized = datetime.now(UTC)
        duration = self.asset_run.materialized - self.asset_run.runtime
        self.asset_run.duration_miliseconds = int(duration.total_seconds() * 1000)

        await self.asset_run.persist()

        print(
            f"Completed operations for asset_run_id: {self.asset_run.id}, on asset_id: {self.id}, on asset: {self.name}"
        )

        return self.result_artifacts[0]

    def _create_result_artifacts(self) -> List[DataArtifact]:
        base_path = self.path.split(".")[0]
        result_artifacts = []
        for filetype in self.result_artifact_filetypes:
            result_artifacts.append(
                DataArtifact(
                    f"{base_path}.{filetype}",
                    read_json_options=self.read_json_options,
                    read_csv_options=self.read_csv_options,
                )
            )

        return result_artifacts

    async def query(self, query_str: str | None = None):
        result_artifact = await self()

        artifact_query = DataArtifactQuery([result_artifact])
        return await artifact_query.query(query_str)

    def _get_artifact_base_path(self):
        partition = ""

        # If we snapshot artifacts, encapsulate the file in a directory with the runtime= parameter
        # so you can view changes over time
        if self.snapshot_artifacts and self.asset_run.runtime:
            runtime_str = str(self.asset_run.runtime.isoformat()).replace(":", "_")
            partition = (
                f"year={self.asset_run.runtime.year}/month={self.asset_run.runtime.month}/day={self.asset_run.runtime.day}/runtime={runtime_str}/"
                if self.snapshot_artifacts
                else ""
            )

        # Extract folder path for folder set up
        folder_path = os.path.dirname(self.path)

        # Set up the base path for artifact storage
        if not self.artifacts_dir:
            base_path: str = f"{folder_path}/_artifacts/asset={self.name}/{partition}"
        else:
            base_path: str = f"{self.artifacts_dir}/{partition}"

        return base_path

    def _resolve_attribute(self, input_str: str | None = None):
        if not input_str or not self._bound_arguments:
            return input_str

        input_str = input_str.format(**self._bound_arguments.arguments)
        return input_str

    def _sanitize_name(self, name: str) -> str:
        # Replace any character that's not alphanumeric, underscore, or hyphen with an underscore
        return re.sub(r"[^A-Za-z0-9_\-]", "_", name)

    def _generate_asset_guid(self):
        hash_input = f"{self.name}:{self.path}:{self.artifacts_dir}:{str(self._bound_arguments.arguments) if self._bound_arguments else ''}"
        return hashlib.md5(hash_input.encode()).hexdigest()

    def _generate_asset_iteration_guid(self):
        hash_input = f"{self.name}:{self.path}:{self.artifacts_dir}:{self.asset_run.runtime.isoformat() if self.asset_run.runtime else ''}:{str(self._bound_arguments.arguments) if self._bound_arguments else ''}"
        return hashlib.md5(hash_input.encode()).hexdigest()

    async def _get_asset_metadata(self):
        await register_mad_protocol()
        fs = await get_fs()

        metadata_glob = f"{ASSET_METADATA_LOCATION}/asset_name={self.name}/asset_id={self.id}/**/*.json"

        if fs.glob(metadata_glob):
            return duckdb.query(
                f"SELECT UNNEST(data, max_depth:=2) FROM read_json('mad://{metadata_glob}')"
            )

    async def _get_last_materialized(self):
        asset_metadata = await self._get_asset_metadata()

        if not asset_metadata:
            return

        last_materialized_query = duckdb.query(
            "SELECT max(strptime(materialized, '%Y-%m-%dT%H:%M:%S.%fZ')) FROM asset_metadata"
        ).fetchone()

        if last_materialized_query and last_materialized_query[0]:
            # Convert DuckDB timestamp to Python datetime
            last_materialized = datetime.fromisoformat(str(last_materialized_query[0]))
            # Ensure it's UTC
            return last_materialized.replace(tzinfo=timezone.utc)

    async def _cached_result(self, result_artifact: DataArtifact, runtime: datetime):
        # Check if data has been materialized within cache_expiration period
        if (
            self.last_materialized
            and (self.last_materialized > runtime - self.cache_expiration)
            and await result_artifact.exists()
        ):
            print(
                f"Retrieving cached result_artifact for asset_id: {self.id} | asset: {self.name}"
            )
            return True

    def get_result_artifact_filetypes(self) -> List[str]:
        filetypes_part = self.path.split(".")[-1]
        return filetypes_part.split("|")



import io
from io import StringIO
import os
from typing import Any, cast
from pandas import DataFrame, read_parquet, read_csv
import prefect.filesystems
import prefect.utilities.asyncutils
import fsspec
import fsspec.utils
from prefect.serializers import JSONSerializer
from prefect.blocks.fields import SecretDict
from pydantic import model_validator
import sshfs
import tempfile


fsspec.register_implementation("ssh", sshfs.SSHFileSystem)
fsspec.register_implementation("sftp", sshfs.SSHFileSystem)

FILESYSTEM_URL = os.getenv("FILESYSTEM_URL", "file://./.tmp/storage")
FILESYSTEM_BLOCK_NAME = os.getenv("FILESYSTEM_BLOCK_NAME")


class FsspecFileSystem(
    prefect.filesystems.WritableFileSystem,
    prefect.filesystems.WritableDeploymentStorage,
):
    _block_type_name = "Fsspec Advanced FileSystem"

    basepath: str
    storage_options: SecretDict

    def __init__(
        self,
        basepath: str,
        storage_options: SecretDict | dict | None = None,
        **kwargs,
    ):
        if isinstance(storage_options, dict):
            storage_options = SecretDict(storage_options)

        storage_options = storage_options or SecretDict(kwargs or {})

        super().__init__(
            basepath=basepath,
            storage_options=storage_options,
            **kwargs,
        )

        fs, fs_url = fsspec.core.url_to_fs(
            basepath,
            **(storage_options.get_secret_value()),
        )

        self._fs: fsspec.AbstractFileSystem = fs
        self._fs_url: str = fs_url

    def _resolve_path(self, path: str):
        # resolve the path relative to the basepath as supplied by fsspec
        return f"{self._fs_url.rstrip('/')}/{path.lstrip('/')}"

    def glob(self, path: str):
        # return relative paths to the basepath
        abs_paths = self._fs.glob(self._resolve_path(path))

        return [
            cast(str, abs_path).replace(f"{self._fs_url}/", "")
            for abs_path in abs_paths
        ]

    def mkdirs(self, path: str, exist_ok: bool = False):
        self._fs.mkdirs(self._resolve_path(path), exist_ok=exist_ok)

    def exists(self, path: str):
        return self._fs.exists(self._resolve_path(path))

    async def read_path(self, path: str) -> bytes:
        path = self._resolve_path(path)

        # Check if the path exists
        if not self._fs.exists(path):
            raise ValueError(f"Path {path} does not exist.")

        # Validate that its a file
        if self._fs.info(path)["type"] != "file":
            raise ValueError(f"Path {path} is not a file.")

        self._fs.mkdirs(self._fs._parent(path), exist_ok=True)
        file = self._fs.read_bytes(path)

        return cast(bytes, file)

    async def write_path(self, path: str, content: bytes):
        resolved_path = self._resolve_path(path)

        # Ensure the directory is created
        self._fs.mkdirs(self._fs._parent(resolved_path), exist_ok=True)

        # Create a temporary file
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            temp_file.write(content)
            temp_file.flush()

            # Remember the temporary file name to delete it later
            temp_path = temp_file.name

        # Upload the temporary file to the destination path
        try:
            self._fs.put(temp_path, resolved_path)
        finally:
            # Ensure the temporary file is deleted
            os.remove(temp_path)

        return path

    async def move_path(self, path: str, dest: str) -> str:
        source_path = self._resolve_path(path)
        destination_path = self._resolve_path(dest)

        # Ensure source path exists
        if not self._fs.exists(source_path):
            raise ValueError(f"Source path {source_path} does not exist.")

        # Ensure destination's parent directory exists
        self._fs.mkdirs(self._fs._parent(destination_path), exist_ok=True)

        # Move the file
        self._fs.move(source_path, destination_path)

        return destination_path

    async def delete_path(self, path: str, recursive: bool = False) -> None:
        resolved_path = self._resolve_path(path)

        if not self.exists(path):
            return

        self._fs.rm(resolved_path, recursive=recursive)

    async def get_directory(
        self, from_path: str | None = None, local_path: str | None = None
    ) -> None:
        raise NotImplementedError()

    async def put_directory(
        self,
        local_path: str | None = None,
        to_path: str | None = None,
        ignore_file: str | None = None,
    ) -> None:
        raise NotImplementedError()

    async def write_data(
        self,
        path: str,
        data: list | dict | Any,
        indent: bool = True,
        **kwargs,
    ):
        if isinstance(data, dict):
            # if path has variables, substitute them for the values inside data
            # but only if the data is a simple dict
            path = path.format(**{**kwargs, **data})
        elif isinstance(data, list):
            path = path.format(**{**kwargs, "data": data})

        # infer the serialization type from the path

        if path.lower().endswith(".parquet"):
            buf = io.BytesIO()
            DataFrame(data).to_parquet(buf)
            buf.seek(0)
            data = buf.getvalue()

        elif path.lower().endswith(".csv"):
            buf = StringIO()
            DataFrame(data).to_csv(buf, encoding="utf-8", sep=",")
            data = buf.getvalue().encode("utf-8")

        # otherwise just write using json as default
        else:
            js = (
                JSONSerializer(dumps_kwargs={"indent": 4})
                if indent
                else JSONSerializer()
            )
            data = js.dumps(data)

        # write to the fs
        return await self.write_path(path, data)

    async def read_data(self, path: str, encoding: str = "utf-8"):
        data = await self.read_path(path)

        # infer the deserialization type from the path
        if path.lower().endswith(".parquet"):
            buf = io.BytesIO(data)
            data = read_parquet(buf).to_dict(orient="records")

        elif path.lower().endswith(".csv"):
            # Decode bytes to string using UTF-8
            file_str = data.decode(encoding)
            buf = StringIO(file_str)
            # Parse the string as CSV
            data = read_csv(buf, sep=",").to_dict(orient="records")

        # otherwise assume json as default
        else:
            js = JSONSerializer()
            data = js.loads(data)

        return data

    async def open(self, path: str, mode: str = "rb", auto_mkdir: bool = False):
        resolved_path = self._resolve_path(path)

        if "w" in mode:
            if auto_mkdir:
                self._fs.mkdirs(self._fs._parent(resolved_path), exist_ok=True)
        else:
            # Check if the path exists and is a file
            if not self._fs.exists(resolved_path):
                raise ValueError(f"Path {resolved_path} does not exist.")
            if self._fs.info(resolved_path)["type"] != "file":
                raise ValueError(f"Path {resolved_path} is not a file.")

        # Open the file and read the contents
        return self._fs.open(resolved_path, mode=mode)


_get_fs_result: FsspecFileSystem | None = None


async def get_fs():
    global _get_fs_result

    if not _get_fs_result:
        if FILESYSTEM_BLOCK_NAME:
            _get_fs_result = cast(FsspecFileSystem, await FsspecFileSystem.load(FILESYSTEM_BLOCK_NAME))  # type: ignore
        else:
            _get_fs_result = FsspecFileSystem(basepath=FILESYSTEM_URL)

    return _get_fs_result



from typing import cast
import duckdb
import fsspec
from mad_prefect.filesystems import get_fs
from fsspec.implementations.dirfs import DirFileSystem


class MadFileSystem(DirFileSystem):
    protocol = "mad"

    def __init__(self, basepath: str, storage_options: dict | None = None, **kwargs):
        options = storage_options or kwargs
        fs, fs_url = cast(
            tuple[fsspec.AbstractFileSystem, str],
            fsspec.core.url_to_fs(basepath, **options),
        )

        super().__init__(path=fs_url.rstrip("/"), fs=fs)


async def register_mad_protocol(connection: duckdb.DuckDBPyConnection | None = None):
    if connection and connection.filesystem_is_registered("mad"):
        return
    elif not connection and duckdb.filesystem_is_registered("mad"):
        return

    fs = await get_fs()
    mad_fs = MadFileSystem(fs.basepath, fs.storage_options.get_secret_value())

    if connection:
        connection.register_filesystem(cast(str, mad_fs))
    else:
        duckdb.register_filesystem(cast(str, mad_fs))



import datetime
import os
from typing import Callable, Literal

from mad_prefect.data_assets.options import ReadJsonOptions, ReadCSVOptions

ASSET_METADATA_LOCATION = os.getenv("ASSET_METADATA_LOCATION", "_asset_metadata")
ARTIFACT_FILE_TYPES = Literal["parquet", "json", "csv"]


def asset(
    path: str,
    artifacts_dir: str = "",
    name: str | None = None,
    snapshot_artifacts: bool = False,
    artifact_filetype: ARTIFACT_FILE_TYPES = "json",
    read_json_options: ReadJsonOptions | None = None,
    read_csv_options: ReadCSVOptions | None = None,
    cache_expiration: datetime.timedelta | None = None,
):
    # Prevent a circular reference as it references the env variable
    from mad_prefect.data_assets.data_asset import DataAsset

    def decorator(fn: Callable):
        return DataAsset(
            fn,
            path,
            artifacts_dir,
            name,
            snapshot_artifacts,
            artifact_filetype,
            read_json_options,
            read_csv_options,
            cache_expiration,
        )

    return decorator



import json
import os
from typing import BinaryIO, Sequence, cast
import duckdb
import httpx
import jsonlines
import pandas as pd
from mad_prefect.data_assets import ARTIFACT_FILE_TYPES
from mad_prefect.data_assets.options import ReadCSVOptions, ReadJsonOptions
from mad_prefect.data_assets.utils import yield_data_batches
from mad_prefect.duckdb import register_mad_protocol
from mad_prefect.filesystems import get_fs
import pyarrow as pa
import pyarrow.parquet as pq
import pyarrow.csv as pacsv
from mad_prefect.json.mad_json_encoder import MADJSONEncoder


class DataArtifact:
    def __init__(
        self,
        path: str,
        data: object | None = None,
        read_json_options: ReadJsonOptions | None = None,
        read_csv_options: ReadCSVOptions | None = None,
    ):
        self.path = path
        filetype = os.path.splitext(self.path)[1].lstrip(".")

        if filetype not in ["json", "parquet", "csv"]:
            raise ValueError(f"Unsupported file type: {filetype}")

        self.filetype: ARTIFACT_FILE_TYPES = cast(ARTIFACT_FILE_TYPES, filetype)
        self.data = data
        self.read_json_options = read_json_options or ReadJsonOptions()
        self.read_csv_options = read_csv_options or ReadCSVOptions()
        self.persisted = False

    async def persist(self):
        # If we've already persisted this artifact this session, don't do anything
        if self.persisted:
            return True

        if not self._truthy(self.data):
            return False

        await register_mad_protocol()

        # preserve_insertion_order to improve memory usage.
        # https://duckdb.org/docs/guides/performance/how_to_tune_workloads.html#the-preserve_insertion_order-option
        # duckdb.query("SET preserve_insertion_order = false;")

        if isinstance(self.data, duckdb.DuckDBPyRelation):
            # There is a bug with fsspec and duckdb see test: test_overwriting_existing_file
            # to work around the bug, instead of setting (use_tmp_file 0) in the query, we will directly reference
            # the wrapped filesystem
            fs = await get_fs()
            protocol = (
                fs._fs.protocol
                if isinstance(fs._fs.protocol, str)
                else fs._fs.protocol[0]
            )

            # Ensure the directory for the path exists
            fs.mkdirs(os.path.dirname(self.path), exist_ok=True)

            path = fs._resolve_path(self.path)
            d = self.data

            duckdb.register_filesystem(cast(str, fs._fs))
            duckdb.execute(f"COPY d TO '{protocol}://{path}'")
        else:
            if self.filetype == "json":
                await self._persist_json()
            elif self.filetype == "parquet":
                await self._persist_parquet()
            elif self.filetype == "csv":
                await self._persist_csv()
            else:
                raise ValueError(f"Unsupported file format {self.filetype}")

        self.persisted = await self.exists()

        # Release the reference to data to free up memory
        self.data = None

        return self.persisted

    async def _open(self):
        fs = await get_fs()
        return cast(BinaryIO, await fs.open(self.path, "wb", True))

    async def _persist_json(self):
        entities = self._yield_entities_to_persist()
        file: BinaryIO | None = None
        writer: jsonlines.Writer | None = None

        try:
            next_entity = await anext(entities)

            while self._truthy(next_entity):
                # Use the first entity to determine the file's schema
                if not file or not writer:
                    file = await self._open()
                    writer = jsonlines.Writer(
                        file,
                        dumps=lambda obj: json.dumps(obj, cls=MADJSONEncoder),  # type: ignore
                    )

                table_or_batch: pa.RecordBatch | pa.Table = (
                    next_entity
                    if isinstance(next_entity, (pa.Table, pa.RecordBatch))
                    else None
                )

                if table_or_batch:
                    next_entity = table_or_batch.to_pylist()

                if not isinstance(next_entity, Sequence):
                    next_entity = [next_entity]

                writer.write_all(next_entity)
                next_entity = await anext(entities)
        except StopAsyncIteration as e:
            pass
        except Exception as e:
            raise
        finally:
            if writer:
                writer.close()

            if file:
                file.close()

            await entities.aclose()

    async def _persist_parquet(self):
        def __sanitize_data(data):
            """
            Recursively go through the data and replace any empty dictionaries
            with None or a dummy value to avoid Parquet serialization errors.
            """
            if isinstance(data, dict):
                if not data:  # it's an empty dict
                    return None  # or return {'dummy_field': None} to keep the key with a dummy field
                else:
                    return {key: __sanitize_data(value) for key, value in data.items()}
            elif isinstance(data, list):
                return [__sanitize_data(item) for item in data]
            else:
                return data

        entities = self._yield_entities_to_persist()
        file: BinaryIO | None = None
        writer: pq.ParquetWriter | None = None

        try:
            next_entity = await anext(entities)

            while self._truthy(next_entity):
                b = __sanitize_data(next_entity)
                table_or_batch: pa.RecordBatch | pa.Table = (
                    b
                    if isinstance(b, (pa.Table, pa.RecordBatch))
                    else pa.RecordBatch.from_pylist(b)
                )

                # Use the first entity to determine the file's schema
                if not file or not writer:
                    file = await self._open()
                    writer = pq.ParquetWriter(file, table_or_batch.schema)
                else:
                    # If schema has evolved, adjust the current RecordBatch
                    if not table_or_batch.schema.equals(writer.schema):
                        unified_schema = pa.unify_schemas(
                            [writer.schema, table_or_batch.schema],
                            promote_options="permissive",
                        )

                        # Align the RecordBatch with the unified schema
                        table_or_batch = table_or_batch.cast(unified_schema)

                        # Manually adjust the schema of the writer if needed
                        writer.schema = unified_schema

                writer.write(table_or_batch)
                next_entity = await anext(entities)
        except StopAsyncIteration as e:
            pass
        except Exception as e:
            raise
        finally:
            if writer:
                writer.close()

            if file:
                file.close()

            await entities.aclose()

    async def _persist_csv(self):
        entities = self._yield_entities_to_persist()
        file: BinaryIO | None = None
        first_chunk = True  # Track whether we need to write CSV headers

        try:
            next_entity = await anext(entities)

            while self._truthy(next_entity):
                # If it's already a pa.Table or pa.RecordBatch, use it.
                # Otherwise, convert it to a Table from a list-of-dicts or list-of-rows.
                if not isinstance(next_entity, (pa.Table, pa.RecordBatch)):
                    next_entity = pa.Table.from_pylist(next_entity)

                # If the file isn't open yet, open it
                if not file:
                    file = await self._open()

                # Write the current batch to CSV
                pacsv.write_csv(
                    next_entity,
                    file,
                    write_options=pacsv.WriteOptions(include_header=first_chunk),
                )
                first_chunk = False

                next_entity = await anext(entities)

        except StopAsyncIteration:
            pass
        finally:
            if file:
                file.close()
            await entities.aclose()

    async def _yield_entities_to_persist(self):
        from mad_prefect.data_assets.data_asset import DataAsset

        async for batch_data in yield_data_batches(self.data):
            # If the data is an asset, execute it to get the result artifact
            if isinstance(batch_data, DataAsset):
                batch_data = await batch_data()

            # If the entity is a DataAsset, turn it into a DuckDbPyRelation, so it can be handled
            if isinstance(batch_data, DataArtifact):
                # An artifact may not exist for example when there were no results
                if not await batch_data.exists():
                    continue

                batch_data = await batch_data.query()

            if isinstance(batch_data, pd.DataFrame):
                batch_data = duckdb.from_df(batch_data)

            if isinstance(batch_data, (duckdb.DuckDBPyRelation)):
                # Convert duckdb into batches of arrow tables
                reader = batch_data.fetch_arrow_reader(1000)

                try:
                    while True:
                        try:
                            # this will yield a pyarrow RecordBatch
                            batch = reader.read_next_batch()
                            yield batch
                        except StopIteration as stop:
                            break
                finally:
                    reader.close()

            elif isinstance(batch_data, httpx.Response):
                yield batch_data.json()
            else:
                yield batch_data

    async def query(self, query_str: str | None = None):
        from mad_prefect.data_assets.data_artifact_query import DataArtifactQuery

        artifact_query = DataArtifactQuery(
            [self], self.read_json_options, self.read_csv_options
        )
        return await artifact_query.query(query_str)

    async def exists(self):
        fs = await get_fs()
        self.persisted = fs.exists(self.path)
        return self.persisted

    def _truthy(self, data):
        if isinstance(data, pd.DataFrame):
            if data.empty:
                return False
        # duckdb hangs with the not self.data check, so make sure self.data isn't
        # a duckdb pyrelation before checking self.data
        elif isinstance(data, duckdb.DuckDBPyRelation):
            pass
        elif not data:
            return False

        return True



import datetime
import decimal
import json
import uuid


class MADJSONEncoder(json.JSONEncoder):
    def default(self, data):
        if isinstance(data, decimal.Decimal):
            return float(data)

        if isinstance(data, datetime.datetime):
            # DuckDB doesn't recognize the ISO 8601 timestamp format with 'T' by default.
            # DuckDB's automatic type inference for timestamps expects a space ' ' between the date and time components.
            return data.isoformat(
                " ", "microseconds"
            )  # Always serialize microseconds as duckdb does not support a variable date format

        if isinstance(data, datetime.time):
            return data.isoformat()

        if isinstance(data, datetime.date):
            return data.isoformat()

        if isinstance(data, uuid.UUID):
            return str(data)

        return super().default(data)



from mad_prefect.data_assets import ARTIFACT_FILE_TYPES
from mad_prefect.data_assets.options import ReadCSVOptions, ReadJsonOptions
from mad_prefect.data_assets.utils import yield_data_batches
from mad_prefect.data_assets.data_artifact import DataArtifact


class DataArtifactCollector:

    def __init__(
        self,
        collector: object,
        dir: str,
        filetype: ARTIFACT_FILE_TYPES = "json",
        artifacts: list[DataArtifact] | None = None,
        read_json_options: ReadJsonOptions | None = None,
        read_csv_options: ReadCSVOptions | None = None,
    ):
        self.collector = collector
        self.dir = dir
        self.filetype = filetype
        self.artifacts = artifacts or []
        self.read_json_options = read_json_options or ReadJsonOptions()
        self.read_csv_options = read_csv_options or ReadCSVOptions()

    async def collect(self):
        fragment_num = 0

        async for fragment in yield_data_batches(self.collector):
            # If the output isn't a DataArtifact manually set the params & base_path
            # and initialize the output as a DataArtifact
            if isinstance(fragment, DataArtifact):
                fragment_artifact = fragment
            else:
                path = self._build_artifact_path(self.dir, fragment_number=fragment_num)
                fragment_artifact = DataArtifact(
                    path, fragment, self.read_json_options, self.read_csv_options
                )

            if await fragment_artifact.persist():
                self.artifacts.append(fragment_artifact)
                fragment_num += 1

        return self.artifacts

    def _build_artifact_path(
        self,
        base_path: str,
        params: dict | None = None,
        fragment_number: int | None = None,
    ):
        filetype = self.filetype
        base_path = base_path.rstrip("/")

        if params is None:
            return f"{base_path}/fragment={fragment_number}.{filetype}"

        params_path = "/".join(f"{key}={value}" for key, value in params.items())

        return f"{base_path}/{params_path}.{filetype}"
